unit BBSUtils;

interface

uses Windows, SysUtils, Classes, IniFiles, Contnrs, StrUtils, HashList;

type
	EBBSException = class(Exception);

	{ T2chCustomIniFile }
	T2chCustomIniFile = class(TPersistent)
	private
		FInternalIniFile: TMemIniFile;
	protected
		procedure DoUpdate(IniFile: TMemIniFile; Direction: Boolean); virtual; abstract;
	public
		constructor Create;
		destructor Destroy; override;
		procedure LoadFromFile(const FileName: string); virtual;
		procedure SaveToFile(const FileName: string); virtual;
		procedure Clear; virtual;
	end;

	{ T2chSubject }

	E2chSubjectError = class(EBBSException);
	T2chSubjectItem = record
		ThreadNum: Currency;
		Title: string;
		ResCount: Integer;
		Depth: Integer;
	end;
	P2chSubjectItem = ^T2chSubjectItem;
	T2chSubject = class;
	T2chSubject = class
	private
		InternalList: TStringList;
		FContentDate: TDateTime;	{ 擾 }
		function GetResCount(Index: Integer): Integer;
    function GetStrings(Index: Integer): string;
    function Get(Index: Integer): P2chSubjectItem;
    function GetCount: Integer;
		function GetDepth(Index: Integer): Integer;
    function GetSorted: Boolean;
    procedure SetSorted(const Value: Boolean);
    function GetLine(Index: Integer): string;
	public
		constructor Create;
		destructor Destroy; override;
		procedure Clear;
		function Add(const ThreadNum, Title: string; ResCount, Depth: Integer): Integer; overload;
		function Add(const ALine: string; CommaSeparated: Boolean = False): Integer; overload;
		procedure LoadFromFile(const FileName: string); virtual;
		procedure LoadFromStream(Stream: TStream); virtual;
		procedure SaveToFile(const FileName: string); virtual;
		procedure SaveToStream(Stream: TStream); virtual;
		function IndexOf(const S: string): Integer; overload;
		function IndexOf(const S: Currency): Integer; overload;
		property Count: Integer read GetCount;
		property Depth[Index: Integer]: Integer read GetDepth;
		property ResCount[Index: Integer]: Integer read GetResCount;
		property Strings[Index: Integer]: string read GetStrings;
		property Items[Index: Integer]: P2chSubjectItem read Get; default;
		property Lines[Index: Integer]: string read GetLine;
		property ContentDate: TDateTime read FContentDate;
		property Sorted: Boolean read GetSorted write SetSorted;
	end;
	T2chSubjectIniFile = class(T2chCustomIniFile)
  private
    FLastModified: string;
    FETag: string;
	protected
		procedure DoUpdate(IniFile: TMemIniFile; Direction: Boolean); override;
	public
		procedure Clear; override;
		property ETag: string read FETag write FETag;
		property LastModified: string read FLastModified write FLastModified;
	end;

	{ T2chThreadDat }

  E2chThreadDatError = class(EBBSException);
	T2chThreadDat = class(TStringList)
	private
		FCommaSeparated: Boolean;
		FModified: Boolean;
    function GetDate(Index: Integer): string;
    function GetDateTime(Index: Integer): TDateTime;
    function GetMail(Index: Integer): string;
    function GetName(Index: Integer): string;
    function GetMessage(Index: Integer): string;
	public
		procedure Aborn(Index: Integer; Invisible: Boolean);
		function IdAborn(const IdStr: string; const Invisible: Boolean): Integer;
		procedure LoadFromStream(Stream: TStream); override;
		function ToHTML(Index, Count: Integer): string;
		procedure Split(Index: Integer; var StrArray: array of string);
		property Dates[Index: Integer]: string read GetDate;
		property DateTimes[Index: Integer]: TDateTime read GetDateTime;
		property Mails[Index: Integer]: string read GetMail;
		property Messages[Index: Integer]: string read GetMessage;
		property Names[Index: Integer]: string read GetName;
		property Modified: Boolean read FModified;
		property CommaSeparated: Boolean read FCommaSeparated write FCommaSeparated;
	end;

	{ T2chThreadIniFile }

	T2chThreadIniResState = (rsNone, rsLoading, rsNewThread, rsNew, rsFreezed, rsError);
	T2chThreadIniFile = class(T2chCustomIniFile)
	private
		FLinked: Boolean;	{ t@CƃNĂ邩ǂ }
		FModified: Boolean;
    FThreadNum: Currency;
    FNewResCount: Integer;
    FDatSize: Integer;
    FRealResCount: Integer;
    FPriority: Integer;
    FResCount: Integer;
    FServerName: string;
    FBoardPath: string;
    FMemo: string;
    FDateTime: TDateTime;
    FWriteTime: TDateTime;
    FThreadTitle: string;
    FDepth: Integer;
    FBoardName: string;
    FListItem: Pointer;
    FSortIndex: Integer;
    FResState: T2chThreadIniResState;
    FViewRes: Integer;
    FFreezed: Boolean;
		procedure SetBoardPath(const Value: string);
    procedure SetDateTime(const Value: TDateTime);
    procedure SetDatSize(const Value: Integer);
    procedure SetPriority(const Value: Integer);
    procedure SetMemo(const Value: string);
    procedure SetNewResCount(const Value: Integer);
    procedure SetRealResCount(const Value: Integer);
    procedure SetResCount(const Value: Integer);
    procedure SetServerName(const Value: string);
    procedure SetThreadNum(const Value: Currency);
    procedure SetWriteTime(const Value: TDateTime);
    procedure SetThreadTitle(const Value: string);
    function GetLogPath: string;
    procedure SetBoardName(const Value: string);
    procedure SetViewRes(const Value: Integer);
		procedure SetFreezed(const Value: Boolean);
	protected
		procedure DoUpdate(IniFile: TMemIniFile; Direction: Boolean); override;
	public
		constructor Create;
		procedure LoadFromFile(const FileName: string); override;
		procedure SaveToFile(const FileName: string); override;
		property ThreadNum: Currency read FThreadNum write SetThreadNum;
		property RealResCount: Integer read FRealResCount write SetRealResCount;
		property ResCount: Integer read FResCount write SetResCount;
		property NewResCount: Integer read FNewResCount write SetNewResCount;
		property ServerName: string read FServerName write SetServerName;
		property BoardPath: string read FBoardPath write SetBoardPath;
		property BoardName: string read FBoardName write SetBoardName;
		property DateTime: TDateTime read FDateTime write SetDateTime;
		property WriteTime: TDateTime read FWriteTime write SetWriteTime;
		property Priority: Integer read FPriority write SetPriority;
		property Memo: string read FMemo write SetMemo;
		property DatSize: Integer read FDatSize write SetDatSize;
		property ThreadTitle: string read FThreadTitle write SetThreadTitle;
		property ViewRes: Integer read FViewRes write SetViewRes;
		property Linked: Boolean read FLinked write FLinked;
		property Depth: Integer read FDepth write FDepth;
		property Freezed: Boolean read FFreezed write SetFreezed;
		property LogPath: string read GetLogPath;
		property ListItem: Pointer read FListItem write FListItem;
		property SortIndex: Integer read FSortIndex write FSortIndex;
		property ResState: T2chThreadIniResState read FResState write FResState;
		property Modified: Boolean read FModified;
	end;

	{ T2chIndexList }

	T2chIndexList = class(TObjectList)
	private
		FHashList: THashList;
    FModified: Boolean;
		function GetIndexItem(Index: Integer): T2chThreadIniFile;
		procedure SetIndexItem(Index: Integer; const Value: T2chThreadIniFile);
	protected
		procedure Notify(Ptr: Pointer; Action: TListNotification); override;
	public
		constructor Create(AOwnsObjects: Boolean);
		destructor Destroy; override;
		procedure CreateStorageList(const LogDir, ServerName, BoardPath: string);
		function LogPathOfObject(const LogPath: string): T2chThreadIniFile;
		property Modified: Boolean read FModified;
		property Items[Index: Integer]: T2chThreadIniFile read GetIndexItem write SetIndexItem; default;
	end;

function HTML2String(const HTML: string; IgnoreTag: Boolean = False): string;
function StringSplit(Text: string; const SubStr: string; var List: array of string): Integer;
function URLToPath(const URL: string): string;
procedure ExtractServerNameAndBoardPath(const Path: string; var ServerName, BoardPath: string);
function MyStrToDateTime(const S: string): TDateTime;
function MyDateTimeToStr(const DateTime: TDateTime): string;

const
	BOARD_2CHANNEL_BRD = '2channel.brd';
	BOARD_OTHER_BRD = 'other.brd';

implementation

uses regex, AWKFunc;

{ HTML2String }
function HTML2String(const HTML: string; IgnoreTag: Boolean): string;
var
	Text, S: string;
	I, J: Integer;
begin
  Result := '';

	if IgnoreTag then
		Text := HTML
	else begin
		Text := AWKGSub('<[Bb][Rr]>', #13#10, HTML);
		Text := AWKGSub('<[^>]+>', '', Text);
		Text := AWKGSub('[ '#9']+', ' ', Text);
	end;
	
	I := 1;

	while I <= Length(Text) do
	begin
		if Text[I] = '&' then
		begin
			S := '';

			J := I + 1;
			while J <= Length(Text) do
			begin
				case Text[J] of
					'A'..'Z', 'a'..'z': begin
						S := S + Text[J];
						Inc(J);
					end;
					else Break;
				end;
			end;

			if J > Length(Text) then
			begin
				Result := Result + Copy(Text, I, MaxInt);
				Break;
			end;
			
			if Text[J] = ';' then
			begin
				if S = 'amp' then
					Result := Result + '&'
				else if S = 'gt' then
					Result := Result + '>'
				else if S = 'lt' then
					Result := Result + '<'
				else if (S = 'nbsp') or (S = 'ensp') or (S = 'thinsp') then
					Result := Result + ' '
				else if S = 'copy' then
					Result := Result + '(c)'
				else
					Result := Result + '&' + S + ';';
				I := J + 1;
			end else begin
				Result := Result + Copy(Text, I, J - I + 1);
				I := J + 1;
			end;
		end else begin
			Result := Result + Text[I];
			Inc(I);
		end;
	end;
end;

{CommaSplit}
function CommaSplit(var List: array of string; const str: string): Integer;
var
	I, Count: Integer;
	TmpStr: string;
begin
	TmpStr := '';
	Count := 0;

	for I := 1 to Length(str) do
	begin
		if str[I] = ',' then
		begin
			TmpStr := StringReplace(TmpStr, 'M', ',', [rfReplaceAll]);
			List[Count] := TmpStr;
			Inc(Count);
			TmpStr := '';
		end else
			TmpStr := TmpStr + str[I];
	end;

	if TmpStr <> '' then
	begin
		List[Count] := TmpStr;
		Inc(Count);
	end;

	Result := Count;
end;

{ StringSplit }
function StringSplit(Text: string; const SubStr: string; var List: array of string): Integer;
var
	I, Index, Count, Len: Integer;
begin
	Index := Low(List);
	Count := 0;
	Len := Length(SubStr);
	
	I := Pos(SubStr, Text);
	if I = 0 then
	begin
		List[Index] := Text;
		Result := 1;
		Exit;
	end;

	repeat
		List[Index] := Copy(Text, 1, I - 1);
		if Length(Text) < I + Len then
		begin
			Text := '';
			Break;
		end;
		Text := Copy(Text, I + Len, MaxInt);

		Inc(Index);
		Inc(Count);

		I := Pos(SubStr, Text);
	until I = 0;

	if Text <> '' then
	begin
		Inc(Count);
		List[Index] := Text;
	end;

	Result := Count;
end;

function URLToPath(const URL: string): string;
begin
	Result := AnsiReplaceStr(URL, '/', '_');
	Result := AnsiReplaceStr(Result, ':', '_');
end;

procedure ExtractServerNameAndBoardPath(const Path: string; var ServerName, BoardPath: string);
var
	I: Integer;
begin
	I := Length(Path);
	while I > 0 do
	begin
		if Path[I] = '/' then
			Break;
		Dec(I);
	end;
	if I <= 0 then
	begin
		ServerName := '';
		BoardPath := Path;
	end else
	begin
		ServerName := Copy(Path, 1, I - 1);
		BoardPath := Copy(Path, I + 1, MaxInt);
	end;
end;

{ MyStrToDateTime }
function MyStrToDateTime(const S: string): TDateTime;
begin
	{ yyyy/mm/dd hh:nn:ss }
	Result := EncodeDate(StrToInt(Copy(S, 1, 4)), StrToInt(Copy(S, 6, 2)), StrToInt(Copy(S, 9, 2))) +
						EncodeTime(StrToInt(Copy(S, 12, 2)), StrToInt(Copy(S, 15, 2)), StrToInt(Copy(S, 18, 2)), 0);
end;

{ MyDateTimeToStr }
function MyDateTimeToStr(const DateTime: TDateTime): string;
{var
	Year, Month, Day: Word;
	Hour, Min, Sec, MSec: Word;
begin
	DecodeDate(DateTime, Year, Month, Day);
	DecodeTime(DateTime, Hour, Min, Sec, MSec);
	Result := Format('%.4d/%.2d/%.2d %.2d:%.2d:%.2d', [Year, Month, Day, Hour, Min, Sec]);
end;}
const
	DateFmt = 'yyyy/mm/dd hh:nn:ss';
begin
	Result := FormatDateTime(DateFmt, DateTime);
end;

{ T2chSubject }

function T2chSubject.Add(const ThreadNum, Title: string; ResCount,
	Depth: Integer): Integer;
var
	Item: P2chSubjectItem;
{ ZeroFill }
	function ZeroFill(const Str: string): string;
	var
		I: Integer;
	begin
		I := 10 - Length(Str);
		if I > 0
			then Result := StringOfChar('0', I) + Str
			else Result := Str;
	end;
begin
	New(Item);
	Item.ThreadNum := StrToCurr(ThreadNum);
	Item.Title := Title;
	Item.ResCount := ResCount;
	Item.Depth := Depth;
	Result := InternalList.AddObject(ZeroFill(ThreadNum), TObject(Item));
end;

function T2chSubject.Add(const ALine: string; CommaSeparated: Boolean): Integer;
var
	Token: array[0 .. 1] of string;
	P: Integer;
	ThreadNum, ResCount, SubStr: string;
	Item: P2chSubjectItem;
{ Numeric }
	function Numeric(const Text: string): string;
	var
		I, Count: Integer;
	begin
		Count := 0;
		for I := 1 to Length(Text) do
		begin
			if Text[I] in ['0' .. '9'] then
				Inc(Count)
			else
				Break;
		end;

		Result := Copy(Text, 1, Count);
	end;
{ GetLastParenthesis }
	function GetLastParenthesis(const Text: string): Integer;
	var
		I: Integer;
	begin
		Result := -1;
		for I := 1 to Length(Text) do
			if Text[I] = '(' then Result := I;
	end;
{ ZeroFill }
	function ZeroFill(const Str: string): string;
	var
		I: Integer;
	begin
		I := 10 - Length(Str);
		if I > 0
			then Result := StringOfChar('0', I) + Str
			else Result := Str;
	end;
begin
	if CommaSeparated
		then SubStr := ','
		else SubStr := '<>';

	New(Item);
	StringSplit(ALine, SubStr, Token);
	ThreadNum := Numeric(Token[0]);
	P := GetLastParenthesis(Token[1]);
	ResCount := Numeric(Copy(Token[1], P + 1, MaxInt));
	Dec(P);
	if (not CommaSeparated) and (Token[1][P] = ' ') then
		Dec(P);
	Item.ThreadNum := StrToCurr(ThreadNum);
	Item.ResCount := StrToInt(ResCount);
	Item.Title := Copy(Token[1], 1, P);
	if CommaSeparated then
		Item.Title := StringReplace(Item.Title, 'M', ',', [rfReplaceAll]);
	Item.Depth := InternalList.AddObject(ZeroFill(ThreadNum), TObject(Item)) + 1;

	Result := Item.Depth - 1;
end;

procedure T2chSubject.Clear;
var
	I: Integer;
begin
	for I := 0 to InternalList.Count -1 do
		Dispose(P2chSubjectItem(InternalList.Objects[I]));

	InternalList.Clear;
end;

constructor T2chSubject.Create;
begin
	inherited;

	InternalList := TStringList.Create;
(*	InternalList.Sorted := True; { \[g͂Ȃ }*)

	FContentDate := -1;
end;

destructor T2chSubject.Destroy;
begin
	Clear;
	InternalList.Free;

  inherited;
end;

function T2chSubject.Get(Index: Integer): P2chSubjectItem;
begin
	Result := P2chSubjectItem(InternalList.Objects[Index]);
end;

function T2chSubject.GetCount: Integer;
begin
	Result := InternalList.Count;
end;

function T2chSubject.GetDepth(Index: Integer): Integer;
begin
	Result := P2chSubjectItem(InternalList.Objects[Index]).Depth;
end;

function T2chSubject.GetLine(Index: Integer): string;
var
	Item: P2chSubjectItem;
begin
	Item := P2chSubjectItem(InternalList.Objects[Index]);
	Result := InternalList[Index] + '.dat<>' + Item.Title + ' (' + IntToStr(Item.ResCount) + ')';
end;

function T2chSubject.GetResCount(Index: Integer): Integer;
begin
	Result := P2chSubjectItem(InternalList.Objects[Index]).ResCount;
end;

function T2chSubject.GetSorted: Boolean;
begin
	Result := InternalList.Sorted;
end;

function T2chSubject.GetStrings(Index: Integer): string;
begin
	Result := InternalList[Index];
end;

function T2chSubject.IndexOf(const S: Currency): Integer;
var
	I: Integer;
	function BinSearch(StartIndex, EndIndex: Integer): Integer;
	var
		M: Integer;
		Value: Currency;
	begin
		if EndIndex < StartIndex then
		begin
			Result := -1;
			Exit;
		end;

		M := StartIndex + (EndIndex - StartIndex) div 2;
		Value := P2chSubjectItem(InternalList.Objects[M]).ThreadNum;

		if S < Value then
			Result := BinSearch(StartIndex, M - 1)
		else if S > Value then
			Result := BinSearch(M + 1, EndIndex)
		else
			Result := M;
	end;
begin
	Result := -1;

	if InternalList.Sorted then
	begin
		Result := BinSearch(0, InternalList.Count - 1);
		Exit;
	end;

	for I := 0 to InternalList.Count - 1 do
		if S = P2chSubjectItem(InternalList.Objects[I]).ThreadNum then
		begin
			Result := I;
			Exit;
		end;
end;

function T2chSubject.IndexOf(const S: string): Integer;
begin
	Result := IndexOf(StrToCurr(S));
end;

procedure T2chSubject.LoadFromFile(const FileName: string);
var
	Stream: TStream;
	AccessTime, LocalAccessTime: FILETIME;
	SysAccessTime: SYSTEMTIME;
begin
	Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
	try
		LoadFromStream(Stream);
		GetFileTime(TFileStream(Stream).Handle, nil, nil, @AccessTime);
		FileTimeToLocalFileTime(AccessTime, LocalAccessTime);
		FileTimeToSystemTime(LocalAccessTime, SysAccessTime);
		FContentDate := SystemTimeToDateTime(SysAccessTime);
	finally
		Stream.Free;
	end;
end;

procedure T2chSubject.LoadFromStream(Stream: TStream);
var
	List: TStringList;
	I: Integer;
	CommaSeparated: Boolean;
begin
	List := nil;

	try
		List := TStringList.Create;

		List.LoadFromStream(Stream);

		CommaSeparated := False;
		{ ܂aaȑꍇ }
		if (List.Count > 0) and (AWKMatch('^[0-9]+\.cgi,', List[0], nil) >= 0) then
			CommaSeparated := True;

		for I := 0 to List.Count - 1 do
			Add(List[I], CommaSeparated);

	finally
		List.Free;
	end;
end;

procedure T2chSubject.SaveToFile(const FileName: string);
var
	Stream: TStream;
begin
	Stream := TFileStream.Create(FileName, fmCreate);
	try
		SaveToStream(Stream);
	finally
		Stream.Free;
	end;
end;

procedure T2chSubject.SaveToStream(Stream: TStream);
var
	List: TStringList;
	Item: P2chSubjectItem;
	I: Integer;
begin
	List := nil;

	try
		List := TStringList.Create;

		for I := 0 to InternalList.Count - 1 do
		begin
			Item := P2chSubjectItem(InternalList.Objects[I]);
			List.Add(InternalList[I] + '.dat<>' + Item.Title + ' (' + IntToStr(Item.ResCount) + ')');
		end;

		List.SaveToStream(Stream);
	finally
		List.Free;
	end;
end;

procedure T2chSubject.SetSorted(const Value: Boolean);
begin
	InternalList.Sorted := Value;
end;

{ T2chThreadDat }

procedure T2chThreadDat.Aborn(Index: Integer; Invisible: Boolean);
begin
{	Put(Index, IfThen(Invisible, KATJUSHA_INVISIBLE_ABORN, KATJUSHA_ABORN));}
	FModified := True;
end;

function T2chThreadDat.GetDateTime(Index: Integer): TDateTime;
var
	DateStr: string;
	MatchIdx, AYear: Integer;
	DateAndTime: array [0 .. 5] of string;
	function IfThen(AValue: Boolean; ATrue: Integer; AFalse: Integer): Integer;
	begin
		if AValue
			then Result := ATrue
			else Result := AFalse;
	end;
begin
	DateStr := GetDate(Index);

	{ 02/05/23 18:39 } 
	MatchIdx := AWKMatch('([0-9]+)/([0-9]+)/([0-9]+)[ ]+([0-9]+):([0-9]+)',
			DateStr, DateAndTime);
	if MatchIdx < 0 then
	begin
		{ 2002/12/29() 05:42 }
		MatchIdx := AWKMatch('([0-9]+)/([0-9]+)/([0-9]+)\([^\)]+\)[ ]+([0-9]+):([0-9]+)',
				DateStr, DateAndTime);
	end;

	if MatchIdx < 0 then
		raise E2chThreadDatError.Create('"' + DateStr + '": t܂');

	AYear := StrToInt(DateAndTime[1]);
	if AYear < 100 then
		AYear := AYear + IfThen(AYear >= 70, 1900, 2000);
	
	Result := EncodeDate(AYear, StrToInt(DateAndTime[2]), StrToInt(DateAndTime[3])) +
						EncodeTime(StrToInt(DateAndTime[4]), StrToInt(DateAndTime[5]), 0, 0);
end;

function T2chThreadDat.GetDate(Index: Integer): string;
var
	DatArray: array [0 .. 4] of string;
begin
	Split(Index, DatArray);
	Result := DatArray[2];
end;

function T2chThreadDat.GetMail(Index: Integer): string;
var
	DatArray: array [0 .. 4] of string;
begin
	Split(Index, DatArray);
	Result := DatArray[1];
end;

function T2chThreadDat.GetMessage(Index: Integer): string;
var
	DatArray: array [0 .. 4] of string;
begin
	Split(Index, DatArray);
	Result := DatArray[3];
end;

function T2chThreadDat.GetName(Index: Integer): string;
var
	DatArray: array [0 .. 4] of string;
begin
	Split(Index, DatArray);
	Result := DatArray[0];
end;

function T2chThreadDat.IdAborn(const IdStr: string;
  const Invisible: Boolean): Integer;
var
	DateStr: string;
	P, I, C: Integer;
	M: Boolean;
begin
	M := False;
	C := 0;

	for I := 0 to Count - 1 do
	begin
		DateStr := GetDate(I);
		if DateStr = 'ځ`' then Continue;

		P := Pos('ID:', DateStr);
		if ((P > 0) and (Copy(DateStr, P + 3, 8) = IdStr)) then
		begin
			Inc(C);
{			Put(I, IfThen(Invisible, KATJUSHA_INVISIBLE_ABORN, KATJUSHA_ABORN));}
			M := True;
		end;
	end;

	if M then
		FModified := True;

	Result := C;
end;

procedure T2chThreadDat.LoadFromStream(Stream: TStream);
var
	S: string;
begin
  inherited;
	FModified := False;
	FCommaSeparated := True;
	try
		S := Get(0);
		if Pos('<>', S) > 0 then
			FCommaSeparated := False;
	except
		{ O͖ }
	end;
end;

procedure T2chThreadDat.Split(Index: Integer;
  var StrArray: array of string);
begin
	if FCommaSeparated then
		CommaSplit(StrArray, Get(Index))
	else
		StringSplit(Get(Index), '<>', StrArray);
end;

function T2chThreadDat.ToHTML(Index, Count: Integer): string;
const
	RES_HTML = 'Res.htm';
	HEADER_HTML = 'Header.htm';
var
	ResText, Text, MailName, Number, DateStr, IdStr: string;
	List: TStringList;
	I, P: Integer;
	DatArray: array [0 .. 4] of string;
begin
	List := TStringList.Create;

	try
		List.LoadFromFile(RES_HTML);
		ResText := List.GetText;
		if Index = 0 then
			List.LoadFromFile(HEADER_HTML)
		else
			List.Clear;

		for I := Index to Index + Count - 1 do
		begin
			Split(I, DatArray);
(*			if DatArray[2] = 'ځ`' then
			begin
				{ AJ[͂Ƃ肠o͂ }
				Text := Format('<a name="a%d"></a>', [I + 1]);
				List.Add(Text);
				Continue;
			end; *)

			Number := Format('<A HREF="menu:%d">%d</A>', [I + 1, I + 1]);
			DateStr := DatArray[2];

			P := Pos('ID:', DateStr);
			if P > 0 then
			begin
				IdStr := Copy(DateStr, P + 3, 8);
				DateStr := StringReplace(DateStr, IdStr, '<A HREF="idmenu:' + IdStr + '">' + IdStr + '</A>', []);
			end;

      DatArray[0] := '<b>' + DatArray[0] + '</b>';

			if DatArray[1] <> '' then
				MailName := '<A HREF="mailto:' + DatArray[1] + '">' + DatArray[0] + '</A>'
			else
				MailName := DatArray[0];

			Text := StringReplace(ResText, '&PLAINNUMBER', IntToStr(I + 1), [rfReplaceAll]);
			Text := StringReplace(Text, '&NUMBER', Number, []);
			Text := StringReplace(Text, '&MAILNAME', MailName, []);
			Text := StringReplace(Text, '&MAIL', DatArray[1], []);
			Text := StringReplace(Text, '&NAME', DatArray[0], []);
			Text := StringReplace(Text, '&DATE', DateStr, []);
			Text := StringReplace(Text, '&MESSAGE', DatArray[3], []);
			Text := Format('<a name="a%d"></a>', [I + 1]) + Text;
			
			List.Add(Text);
		end;

		Result := '';
		
		if List.Count > 0 then
			Result := List.GetText;
	finally
		List.Free;
	end;
end;

{ T2chThreadIniFile }

constructor T2chThreadIniFile.Create;
begin
	inherited;
	
	FLinked := False;
	FDateTime := -1;
	FWriteTime := -1;
	FViewRes := -1;
end;

procedure T2chThreadIniFile.DoUpdate(IniFile: TMemIniFile;
	Direction: Boolean);
const
	ThreadSection = 'Thread';
begin
	if Direction then
	begin
		{ ǂݍ }
		FThreadNum := StrToCurr(IniFile.ReadString(ThreadSection, 'ThreadNum', CurrToStr(FThreadNum)));
		FThreadTitle := IniFile.ReadString(ThreadSection, 'ThreadTitle', FThreadTitle);
		FRealResCount := IniFile.ReadInteger(ThreadSection, 'RealResCount', FRealResCount);
		FResCount := IniFile.ReadInteger(ThreadSection, 'ResCount', FResCount);
		FNewResCount := IniFile.ReadInteger(ThreadSection, 'NewResCount', FNewResCount);
		FServerName := IniFile.ReadString(ThreadSection, 'ServerName', FServerName);
		FBoardPath := IniFile.ReadString(ThreadSection, 'BoardPath', FBoardPath);
		FBoardName := IniFile.ReadString(ThreadSection, 'BoardName', FBoardName);
		FDateTime := MyStrToDateTime(IniFile.ReadString(ThreadSection, 'DateTime', ''));
		FWriteTime := MyStrToDateTime(IniFile.ReadString(ThreadSection, 'WriteTime', ''));
		FPriority := IniFile.ReadInteger(ThreadSection, 'Priority', FPriority);
		FMemo := IniFile.ReadString(ThreadSection, 'Memo', FMemo);
		FDatSize := IniFile.ReadInteger(ThreadSection, 'DatSize', FDatSize);
		FViewRes := IniFile.ReadInteger(ThreadSection, 'ViewRes', FViewRes);
		FFreezed := IniFile.ReadBool(ThreadSection, 'Freezed', FFreezed);

		FLinked := True;
	end
	else
	begin
		{  }
		IniFile.WriteString(ThreadSection, 'ThreadNum', CurrToStr(ThreadNum));
		IniFile.WriteString(ThreadSection, 'ThreadTitle', FThreadTitle);
		IniFile.WriteInteger(ThreadSection, 'RealResCount', FRealResCount);
		IniFile.WriteInteger(ThreadSection, 'ResCount', FResCount);
		IniFile.WriteInteger(ThreadSection, 'NewResCount', FNewResCount);
		IniFile.WriteString(ThreadSection, 'ServerName', FServerName);
		IniFile.WriteString(ThreadSection, 'BoardPath', FBoardPath);
		IniFile.WriteString(ThreadSection, 'BoardName', FBoardName);
		IniFile.WriteString(ThreadSection, 'DateTime', MyDateTimeToStr(FDateTime));
		IniFile.WriteString(ThreadSection, 'WriteTime', MyDateTimeToStr(FWriteTime));
		IniFile.WriteInteger(ThreadSection, 'Priority', FPriority);
		IniFile.WriteString(ThreadSection, 'Memo', FMemo);
		IniFile.WriteInteger(ThreadSection, 'DatSize', FDatSize);
		IniFile.WriteInteger(ThreadSection, 'ViewRes', FViewRes);
		IniFile.WriteBool(ThreadSection, 'Freezed', FFreezed);
	end;
end;

function T2chThreadIniFile.GetLogPath: string;
begin
	Result := Format('%s\%s\%s.ini', [URLToPath(ServerName), BoardPath, FormatCurr('000000000', ThreadNum)]);
end;

procedure T2chThreadIniFile.LoadFromFile(const FileName: string);
begin
	inherited;
	FLinked := True;
	FModified := False;
end;

procedure T2chThreadIniFile.SaveToFile(const FileName: string);
begin
	inherited;
	FModified := False;
end;

procedure T2chThreadIniFile.SetBoardName(const Value: string);
begin
	if FBoardName <> Value then
	begin
		FBoardName := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetBoardPath(const Value: string);
begin
	if FBoardPath <> Value then
	begin
		FBoardPath := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetDateTime(const Value: TDateTime);
begin
	if FDateTime <> Value then
	begin
		FDateTime := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetDatSize(const Value: Integer);
begin
	if FDatSize <> Value then
	begin
		FDatSize := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetPriority(const Value: Integer);
begin
	if FPriority <> Value then
	begin
		FPriority := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetMemo(const Value: string);
begin
	if FMemo <> Value then
	begin
		FMemo := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetNewResCount(const Value: Integer);
begin
	if FNewResCount <> Value then
	begin
		FNewResCount := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetRealResCount(const Value: Integer);
begin
	if FRealResCount <> Value then
	begin
		FRealResCount := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetResCount(const Value: Integer);
begin
	if FResCount <> Value then
	begin
		FResCount := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetServerName(const Value: string);
begin
	if FServerName <> Value then
	begin
		FServerName := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetThreadNum(const Value: Currency);
begin
	if FThreadNum <> Value then
	begin
		FThreadNum := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetThreadTitle(const Value: string);
begin
	if FThreadTitle <> Value then
	begin
		FThreadTitle := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetWriteTime(const Value: TDateTime);
begin
	if FWriteTime <> Value then
	begin
		FWriteTime := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetViewRes(const Value: Integer);
begin
	if FViewRes <> Value then
	begin
		FViewRes := Value;
		FModified := True;
	end;
end;

procedure T2chThreadIniFile.SetFreezed(const Value: Boolean);
begin
	if FFreezed <> Value then
	begin
		FFreezed := Value;
		FModified := True;
	end;
end;

{ T2chIndexList }

constructor T2chIndexList.Create(AOwnsObjects: Boolean);
begin
	inherited;
	FHashList := THashList.Create(300, nil, nil);
end;

procedure T2chIndexList.CreateStorageList(const LogDir, ServerName,
  BoardPath: string);
var
	FindName, LogPath: string;
	FindData: TSearchRec;
	Thread: T2chThreadIniFile;
begin
	LogPath := IncludeTrailingPathDelimiter(LogDir) + URLToPath(ServerName) + '\' + BoardPath;
	FindName := IncludeTrailingPathDelimiter(LogPath) + '*.*';
	
	if FindFirst(FindName, 0, FindData) <> 0 then
		Exit;

	try
		repeat
			if ((FindData.Attr and faDirectory) = 0) and
					(ExtractFileExt(LowerCase(FindData.Name)) = '.ini') and
					(ExtractFileName(LowerCase(FindData.Name)) <> 'subject.ini') then
			begin
				Thread := T2chThreadIniFile.Create;
				Thread.LoadFromFile(IncludeTrailingPathDelimiter(LogPath) + FindData.Name);
				Thread.Depth := -1;
				if Thread.Freezed then Thread.ResState := rsFreezed;
				Add(Thread);
			end;
		until FindNext(FindData) <> 0;
	finally
		FindClose(FindData);
	end;
end;


destructor T2chIndexList.Destroy;
begin
	FHashList.Free;
  inherited;
end;

function T2chIndexList.GetIndexItem(Index: Integer): T2chThreadIniFile;
begin
	Result := T2chThreadIniFile(GetItem(Index));
end;

function T2chIndexList.LogPathOfObject(
	const LogPath: string): T2chThreadIniFile;
var
	ThreadIni: T2chThreadIniFile;
begin
	Result := nil;
	if FHashList.Find(LogPath, ThreadIni) then
		Result := ThreadIni;
end;

procedure T2chIndexList.Notify(Ptr: Pointer; Action: TListNotification);
begin
  inherited;
	case Action of
		lnAdded: FHashList.Add(T2chThreadIniFile(Ptr).GetLogPath, T2chThreadIniFile(Ptr));
		lnExtracted, lnDeleted: FHashList.RemoveData(Ptr);
	end;

	FModified := True;
end;

procedure T2chIndexList.SetIndexItem(Index: Integer;
	const Value: T2chThreadIniFile);
begin
	SetItem(Index, Value);
end;

{ T2chCustomIniFile }

procedure T2chCustomIniFile.Clear;
begin
	if Assigned(FInternalIniFile) then
		FreeAndNil(FInternalIniFile);
end;

constructor T2chCustomIniFile.Create;
begin
	FInternalIniFile := nil;
end;

destructor T2chCustomIniFile.Destroy;
begin
	if Assigned(FInternalIniFile) then
		FreeAndNil(FInternalIniFile);
  inherited;
end;

procedure T2chCustomIniFile.LoadFromFile(const FileName: string);
begin
	if Assigned(FInternalIniFile) then
		FreeAndNil(FInternalIniFile);

	FInternalIniFile := TMemIniFile.Create(FileName);
	DoUpdate(FInternalIniFile, True);
end;

procedure T2chCustomIniFile.SaveToFile(const FileName: string);
var
	List: TStringList;
begin
	List := TStringList.Create;
	try
		if Assigned(FInternalIniFile) then
		begin
			DoUpdate(FInternalIniFile, False);

			if FInternalIniFile.FileName = FileName then
				FInternalIniFile.UpdateFile
			else begin
				FInternalIniFile.GetStrings(List);
				FreeAndNil(FInternalIniFile);
				FInternalIniFile := TMemIniFile.Create(FileName);
				FInternalIniFile.SetStrings(List);
				FInternalIniFile.UpdateFile;
			end;
		end else
		begin
			FInternalIniFile := TMemIniFile.Create(FileName);
			DoUpdate(FInternalIniFile, False);
			FInternalIniFile.UpdateFile;
		end;
	finally
		List.Free;
	end;
end;

{ T2chSubjectIniFile }

procedure T2chSubjectIniFile.Clear;
begin
  inherited;
	FETag := '';
	FLastModified := '';
end;

procedure T2chSubjectIniFile.DoUpdate(IniFile: TMemIniFile;
	Direction: Boolean);
const
	SubjectSection = 'Subject';
begin
	if Direction then
	begin
		{ ǂݍ }
		FETag := IniFile.ReadString(SubjectSection, 'ETag', '');
		FLastModified := IniFile.ReadString(SubjectSection, 'LastModified', '');
	end else begin
		{  }
		IniFile.WriteString(SubjectSection, 'ETag', FETag);
		IniFile.WriteString(SubjectSection, 'LastModified', FLastModified);
	end;
end;

end.

